from WordProcessing.WordDict import *
import os.path
import yaml

__words_file = "words.yaml"


def yaml_compress():
    if os.path.isfile(__words_file):
        saved_words = dict()
        with open(__words_file, mode='r', encoding='utf-8') as wordsin:
            saved_words = yaml.load(wordsin, yaml.Loader)
            hashed = dict()
            hashed_sub = dict()
            for key, value in saved_words.items():
                if type(value) is not dict:
                    continue
                hash_full = hash(str(value))
                if hash_full in hashed:
                    saved_words[key] = saved_words[hashed[hash_full]]
                    continue
                hashed[hash_full] = key

                for subkey, subvalue in value.items():
                    if type(subvalue) is str:
                        continue
                    hash_sub = hash(str(subvalue))
                    if hash_sub in hashed_sub:
                        saved_words[key][subkey] = saved_words[hashed_sub[hash_sub][0]][hashed_sub[hash_sub][1]]
                        continue
                    hashed_sub[hash_sub] = (key, subkey)

        with open(__words_file, mode='w', encoding='utf-8') as wordsout:
            yaml.dump(saved_words, wordsout, indent=2, allow_unicode=True)


def init(db: WordDict):
    if os.path.isfile(__words_file):
        with open(__words_file, mode='r', encoding='utf-8') as wordsin:
            saved_words = yaml.load(wordsin, yaml.Loader)
            if type(saved_words) == dict:
                db.input(saved_words)


def save(db: WordDict, *, to_compress: bool = True):
    with open(__words_file, mode='a', encoding='utf-8') as wordsout:
        cache = db.flush()
        if cache is not None:
            yaml.dump(cache, wordsout, indent=2, allow_unicode=True)
    if to_compress:
        yaml_compress()



from contextlib import suppress


def extract_any(self: list, *args, to_pop=False, if_none=None):
    for to_get in args:
        for from_list in self:
            if to_get == from_list:
                if to_pop:
                    self.remove(from_list)
                return to_get

    return if_none


def extract_all(self: list, *args, to_pop=False):
    to_return = []

    for to_get in args:
        for from_list in self:
            if to_get == from_list:
                if to_pop:
                    self.remove(from_list)
                to_return.append(to_get)

    return to_return


def safe_remove(self: list, elem):
    with suppress(ValueError, AttributeError):
        self.remove(elem)


def rel_format(self: str, mod: str):
    if mod[0] == '+':
        return self + mod[1:]
    if mod[-1] == '+':
        return mod[:-1] + self

    subcount = mod.count('-')
    return self[:-subcount] + mod[subcount:]



from .. import Defines


def declined(word: str, *args, path=(), ask_func):
    return {
        Defines.default: case(word, *args, 'и.п.', path=path + ('Именительный падеж', ), ask_func=ask_func),
        'р.п.': case(word, *args, 'р.п.', path=path + ('Родительный падеж', ), ask_func=ask_func),
        'д.п.': case(word, *args, 'д.п.', path=path + ('Дательный падеж', ), ask_func=ask_func),
        'в.п.': case(word, *args, 'в.п.', path=path + ('Винительный падеж', ), ask_func=ask_func),
        'т.п.': case(word, *args, 'т.п.', path=path + ('Творительный падеж', ), ask_func=ask_func),
        'п.п.': case(word, *args, 'п.п.', path=path + ('Предложный падеж', ), ask_func=ask_func)
    }


def case(word: str, *args, path=(), ask_func):
    if 'ед.ч.' in args or 'мн.ч.' in args:
        return '+' if 'и.п.' in args else ask_func(word, path)

    return {
        Defines.default: '+' if 'и.п.' in args else ask_func(word, path + ('единственное число', )),
        'мн.ч.': ask_func(word, path + ('множественное число', ))
    }



from .FormConstructors import Adj, Noun, Verb
from . import Defines
from typing import Callable


def lookup(self, *args):
    if type(self) is not dict:
        return self

    default_key = Defines.default if Defines.default_key not in self else self[Defines.default_key]

    for arg in args:
        if arg in self:
            return lookup(self[arg], *args)

    value = self[default_key]
    return lookup(value, *args)


def tags(self: dict, word: str):
    if Defines.tags_key in self[word]:
        return self[word][Defines.tags_key]
    return []


def construct_forms(word: str, *args, path=(), ask_func: Callable[[str, list], str]):
    if len(args) == 0:
        return '+'

    origin = {
        Defines.tags_key: list(args)
    }

    if 'гл.' in args:
        return {
            **origin,
            Defines.default: Verb.conjugable(word, *args, path=path, ask_func=ask_func),
            'пов.': Verb.imperative(word, path=path + ('Повелительное наклонение', ), ask_func=ask_func),
            'инф.': '+'
        }
    elif 'сущ.' in args:
        return {
            **origin,
            Defines.default: '+' if 'нескл.' in args else Noun.declined(word, *args, path=path, ask_func=ask_func)
        }
    elif 'пр.' in args:
        return {
            **origin,
            Defines.default: Adj.singular(word, path=path + ('Единственное число', ), ask_func=ask_func),
            'мн.ч.': Adj.plural(word, path=path + ('Множественное число', ), ask_func=ask_func)
        }
    else:
        return {
            **origin,
            Defines.default: '+'
        }



from .Util.Util import extract_all, rel_format
from . import Forms, Defines
from dataclasses import dataclass, field
from typing import Callable
import enum
import random


class Status(enum.Enum):
    OK = enum.auto()
    INVALID_REFERENCE = enum.auto()
    REFERENCE_ALREADY_SET = enum.auto()
    WORD_ALREADY_EXISTS = enum.auto()
    NO_TAGS = enum.auto()
    GOT_PHRASE = enum.auto()
    NOUN_NO_GENDER = enum.auto()
    VERB_NO_PERFECTNESS = enum.auto()
    NO_WORD_FOUND = enum.auto()


@dataclass
class InsertResponse:
    word: str
    tags: list[str] = field(default_factory=list)
    status: Status = Status.OK
    supplement: str = ''


class WordDict():

    __word_types = ('безл.', )
    __word_traits = ('абр', 'пинг', 'перс')

    def __init__(self):
        self.__words = dict()
        self.__cache = []
        self.__confirm_func = Callable[[InsertResponse], bool]
        self.__ask_func = Callable[[str, list], str]

    def set_confirm_func(self, confirm_func=Callable[[InsertResponse], bool]):
        self.__confirm_func = confirm_func

    def set_ask_func(self, ask_func=Callable[[str, list], str]):
        self.__ask_func = ask_func

    def delete(self, *args):
        for name in args:
            del self.__words[name]

    def get(self, *args):
        tags = list(args)

        forms = extract_all(tags, 'ед.ч.', 'мн.ч.', 'и.п.', 'р.п.', 'д.п.', 'в.п.', 'т.п.', 'п.п.', 'п.в.', 'н.в.', 'б.в.', '1л', '2л', '3л', to_pop=True)

        if 'гл.' in args:
            forms.extend(extract_all(tags, 'инф.', 'пов.', 'м.р.', 'с.р.', 'ж.р.', to_pop=True))
        elif 'пр.' in args:
            forms.extend(extract_all(tags, 'одуш.', 'неодуш.', 'м.р.', 'с.р.', 'ж.р.', to_pop=True))

        tag_correlations = tuple(filter(
            lambda word: all(tag in Forms.tags(self.__words, word) for tag in tags), self.__words.keys()
            )) if len(tags) > 0 else tuple(self.__words.keys())

        if len(tag_correlations) == 0:
            return None

        word = random.choice(tag_correlations)

        return rel_format(word, Forms.lookup(self.__words[word], *forms))

    def input(self, extension: dict):
        self.__words.update(extension)

    def flush(self):
        if len(self.__cache) == 0:
            return None

        cached_words = {}
        for key in self.__cache:
            cached_words[key] = self.__words[key]

        self.__cache.clear()

        return cached_words

    def output(self):
        return self.__words

    def insert(self, word: str, tags):
        types = extract_all(tags, *self.__word_types, to_pop=True)
        for word_type in types:
            match word_type:
                case 'безл.':
                    tags.extend(('м.р.', 'ж.р.'))

        traits = extract_all(tags, *self.__word_traits, to_pop=False)
        for word_trait in traits:
            match word_trait:
                case 'абр' | 'пинг':
                    tags.append('нескл.')
                case 'перс':
                    tags.append('ед.ч.')

        tags = list(set(tags))

        reference = None
        for tag in tags:
            if tag[0] == '&':
                reference_to = tag[1:]
                if reference_to not in self.__words:
                    self.__confirm_func(InsertResponse(word, tags, Status.INVALID_REFERENCE, reference_to))
                    return
                else:
                    if reference is not None:
                        self.__confirm_func(InsertResponse(word, tags, Status.REFERENCE_ALREADY_SET, str((reference, reference_to))))
                        return
                    reference = reference_to

        response = InsertResponse(word, tags)
        if word in self.__words:
            response.status = Status.WORD_ALREADY_EXISTS
            response.supplement = str(Forms.tags(self.__words, word))
        elif 'гл.' in tags:
            if 'сов.' not in tags and 'несов.' not in tags:
                response.status = Status.VERB_NO_PERFECTNESS
        elif 'сущ.' in tags:
            if 'м.р.' not in tags and 'с.р.' not in tags and 'ж.р.' not in tags:
                response.status = Status.NOUN_NO_GENDER
        elif len(word.split()) > 1:
            response.status = Status.GOT_PHRASE
        elif len(tags) == 0:
            response.status = Status.NO_TAGS

        to_proceed = self.__confirm_func(response)
        if to_proceed:
            if reference is None:
                self.__words[word] = Forms.construct_forms(word, *tags, ask_func=self.__ask_func)
            else:
                if len(tags) == 1:
                    self.__words[word] = self.__words[reference]
                else:
                    tags.remove('&' + reference)
                    self.__words[word] = dict(self.__words[reference])
                    self.__words[word][Defines.tags_key] = tags
            self.__cache.append(word)



from WordProcessing.WordDict import *
import asyncio


def confirm_word(response: InsertResponse) -> bool:
    print(f'Базовая форма слова: "{response.word}".')
    print('Часть речи: ', end='')
    if 'гл.' in response.tags:
        print('глагол.')
    elif 'пр.' in response.tags:
        print('прилагательное/причастие.')
    elif 'сущ.' in response.tags:
        print('существительное.')
    else:
        print('другое (наречие/деепричастие/...)')
    print(f'Теги: {response.tags}.')

    auto_cancel = True
    match response.status:
        case Status.REFERENCE_ALREADY_SET:
            print(f'!! Нельзя ставить более одного указателя.')
            print(f'   Есть конфликт тегов: {response.supplement}.')
        case Status.INVALID_REFERENCE:
            print(f'!! Тег \'{response.supplement}\' указывает на отсутствующее в словаре слово.')
        case Status.WORD_ALREADY_EXISTS:
            print(f'!! Это слово уже есть в списке.')
            print(f'   Его теги: {response.supplement}.')
        case Status.VERB_NO_PERFECTNESS:
            print('!! Пожалуйста, укажите завершённость глагола: совершённый "сов." или несовершённый "несов."')
        case Status.NO_TAGS:
            print('!! Теги добавляются при вводе. Например, "любить гл. несов."; "человек сущ. м.р."')
        case Status.NOUN_NO_GENDER:
            print('! Вы не указали род существительного (м.р.|с.р.|ж.р.|безл.).')
            print('  Выборка в тексте в основном требует род, что значит, ваше слово практически не будет показываться.')
            auto_cancel = False
        case Status.GOT_PHRASE:
            print('! Крайне не рекомендуем добавлять словосочетания.')
            auto_cancel = False
        case Status.OK:
            auto_cancel = False
        case _:
            print(f'!! Неизвестный статус "{response.status.name}".')
            print(f'   Дополнительная информация: "{response.supplement}".')

    if auto_cancel:
        return False
    print('Добавлять слово? (пустой ввод, если да) ', end='')
    return False if len(input()) != 0 else True


def ask_form(word: str, path=()) -> str:
    print(path[0], end='')
    for dir in path[1:]:
        print(f', {dir}', end='')
    print(f' для слова "{word}": ', end='')
    return input()


async def interact(db: WordDict):
    print('Введите слова:')
    while True:
        word = input()
        if len(word) == 0:
            break
        tokens = word.split()
        word = tokens[0]
        tags = tokens[1:] if len(tokens) > 1 else []

        await asyncio.to_thread(db.insert, word, tags)
        print()

    print('Введите теги:')
    while True:
        tags = input()
        if len(tags) == 0:
            break
        tags_set = set(tags.split())
        print(db.get(*tags_set))
        print()


def init(db: WordDict):
    db.set_confirm_func(confirm_word)
    db.set_ask_func(ask_form)



"""A Russian text generator based on templating texts and word picking"""

__version__ = "0.6.0"

from WordProcessing.WordDict import WordDict
import StdIO.db_yaml as DB
import StdIO.io_shell as IO
import asyncio


async def main():
    word_manager = WordDict()

    DB.init(word_manager)
    IO.init(word_manager)

    try:
        await IO.interact(word_manager)
    finally:
        DB.save(word_manager)


if __name__ == '__main__':
    asyncio.run(main())
